#include <unistd.h>
#include <stdio.h>
#include <fcntl.h>
#include <string.h>
#include <sys/mman.h>
#include <sys/ioctl.h>
#include <iostream>
#include "webcam.h"

using namespace std;
using namespace io;

webcam::webcam(){
	init_ok =    open()
			  && get_device_capabilities()
			  && set_format()
			  && reqest_buffers()
			  && setup_buffer()
			  && activate_streaming();
	if(init_ok == false){
		std::cout << "init error" << std::endl;
	}
	if(param_fmt_jpeg){
	    out_img.type   = abmt::image_type::JPEG;
	}else{
	    out_img.type   = abmt::image_type::YUYV;
	}
	out_img.width  = param_width;
	out_img.height = param_height;
}

bool webcam::io_helper(const char* cmd_name, int cmd, void* data){
	if(ioctl(fd, cmd, data) < 0){
		std::cout << "Webcam io_error for device " <<  param_device  << " cmd:";
		std::cout << cmd_name  << " " << strerror(errno);
		std::cout << std::endl;
		return false;
	}
	return true;
}

bool webcam::open(){
	fd = ::open(param_device.c_str(), O_RDWR | O_NONBLOCK);
	if(fd < 0){
		std::cout << "Error open video device " <<  param_device << std::endl;
		return false;
	}
	return true;
}
bool webcam::get_device_capabilities(){
	return io_helper("VIDIOC_QUERYCAP", VIDIOC_QUERYCAP, &device_capabilities);
}

bool webcam::set_format(){
	struct v4l2_format format;
	format.type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
	if(param_fmt_jpeg){
	    format.fmt.pix.pixelformat = V4L2_PIX_FMT_MJPEG;
	}else{
	    format.fmt.pix.pixelformat = V4L2_PIX_FMT_YUYV;
	}
	format.fmt.pix.width = param_width;
	format.fmt.pix.height = param_height;
	return io_helper("VIDIOC_S_FMT", VIDIOC_S_FMT, &format);
}

bool webcam::reqest_buffers(){
	buf_request.type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
	buf_request.memory = V4L2_MEMORY_MMAP;
	buf_request.count = num_buffers;
	return io_helper("VIDIOC_REQBUFS", VIDIOC_REQBUFS, &buf_request);
}

bool webcam::setup_buffer(){
	for(size_t i = 0; i < num_buffers; ++i){
		bool res = setup_buffer(&buffers[i].cfg, &buffers[i].data, i);
		if(res == false) return false;
	}
	return true;
}

bool webcam::setup_buffer(v4l2_buffer* buff, void** data, int idx){
	memset(buff, 0, sizeof(v4l2_buffer));
	buff->type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
	buff->memory = V4L2_MEMORY_MMAP;
	buff->index = idx;
	if( io_helper("VIDIOC_QUERYBUF", VIDIOC_QUERYBUF, buff) == false){
		return false;
	}

	*data = ::mmap(
		NULL,
		buff->length,
		PROT_READ | PROT_WRITE,
		MAP_SHARED,
		fd,
		buff->m.offset
	);
	if(*data == MAP_FAILED){
		std::cout << "error open video: MAP_FAILED " << strerror(errno) << std::endl;
		return false;
	}
	memset(*data, 0, buff->length);
	return true;
}

bool webcam::activate_streaming(){
	int type = V4L2_BUF_TYPE_VIDEO_CAPTURE; // uint to int
	bool start_ok = io_helper("VIDIOC_STREAMON", VIDIOC_STREAMON, &type);
	if(start_ok == false){
		return false;
	}

	for(size_t i = 0; i < num_buffers -1; ++i){
		// queue all except last
		if(queue_buff(i) == false) return false;
	}

	return true;
}

bool webcam::queue_buff(size_t idx){
	return io_helper("VIDIOC_QBUF", VIDIOC_QBUF, &buffers[idx].cfg);
}

bool webcam::poll() {
	v4l2_buffer rcv;
	static bool slow_down = true;

	memset(&rcv, 0, sizeof(v4l2_buffer));
	rcv.type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
	rcv.memory = V4L2_MEMORY_MMAP;

	bool rcv_new_frame = false;
	
	int io_res  = ioctl(fd, VIDIOC_DQBUF, &rcv);
	if( io_res == 0){
		size_t next_idx_to_queue = rcv.index + num_buffers -1;
		if(next_idx_to_queue >= num_buffers){
			next_idx_to_queue =  next_idx_to_queue - num_buffers;
		}
		bool req_ok = queue_buff(next_idx_to_queue);
		if( req_ok == false ){
			std::cout << "io_error queue buff" << std::endl;
		}
		out_img.data.realloc(rcv.bytesused, buffers[rcv.index].data);
		rcv_new_frame = true;
	}

	if(rcv_new_frame){
	    if(param_half_framerate){
    	    if(slow_down){
	       usleep(1000);
    	       slow_down = false; 
    	    }else{
    	        slow_down = true; 
    	        return true;
    	    }
	    }else{
	        return true;
	    }
	}
	usleep(1000);
	return false;
 }

void webcam::stop_stream(){
	int type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
    if(ioctl(fd, VIDIOC_STREAMOFF, &type) < 0){
    	std::cout << "error open video: VIDIOC_STREAMOFF" << std::endl;
    }
}


webcam::~webcam(){
	stop_stream();
	for (size_t i = 0; i < num_buffers; i++){
		munmap(buffers[i].data, buffers[i].cfg.length);
	}
	close(fd);
}
