/**
 * Author: Hendrik van Arragon, 2022
 * SPDX-License-Identifier: MIT
 */

#include <sched.h>
#include <time.h>
#include <unistd.h>

#include <functional>
#include <memory>
#include <map>
#include <iostream>
#include <fstream>
#include <filesystem>
#include <clocale>

#include <abmt/task.h>
#include <abmt/rt/com.h>
#include <abmt/rt/raster.h>
#include <abmt/serialize_utils.h>
#include <abmt/mutex.h>
#include <abmt/util/bt_exception.h>
#include <abmt/json.h>
#include <abmt/time.h>
#include <abmt/util/str_utils.h>

#include "model_adapter_std.h"
#include "model_com_eio.h"
#include "time_from_rec.h"

using namespace std;
using json = abmt::json;

volatile bool global_stop = false;
model_com_eio* com_ptr = 0;

void on_sig(int sig){
	global_stop = true;
}

void abmt::os::log(std::string s){
	if(com_ptr->tcp_con->connected){
		com_ptr->adapter.log(s);
	}else{
		cout << s << endl;
	}
}
void abmt::os::log_err(std::string s){
	if(com_ptr->tcp_con->connected){
		com_ptr->adapter.log_err(s);
	}else{
		cout << s << endl;
	}
}
void abmt::os::die(std::string s){
	throw abmt::util::bt_exception(s);
}

void abmt::os::die_if(bool condition, std::string msg){
	if(condition){
		abmt::os::die(msg);
	}
}

abmt::time sim_time;
double sim_speed = 0;

abmt::time abmt::time::now(){
	return sim_time;
}

abmt::time get_systemtime(){
	timespec t;
	clock_gettime(CLOCK_REALTIME, &t);
	int64_t res;
	res = t.tv_sec * 1000LL*1000*1000 + t.tv_nsec; // 1000LL very important for 32bit systems
	return abmt::time(res);
}

class sim_task: public abmt::task{
public:
	abmt::rt::raster* r;
	uint32_t raster_index;
	abmt::time start_time;
	uint64_t n_ticks = 0;
	volatile bool execute = false;
	abmt::mutex wait_mutex;

	sim_task(abmt::rt::raster* r, uint32_t r_idx):abmt::task([this]{this->run();},false), r(r), raster_index(r_idx){

	}

	abmt::time next_start_time(){
		return start_time + r->interval*n_ticks;
	}

	void run(){
		try{
			r->init();
			usleep(10*1000); // give async tasks some time to init (see example/io/udp.acfg)
			start_time = abmt::time::now();
			if(r->is_sync == false){
				while(global_stop == false){
					wait_mutex.lock();
					execute = false;
					// unlocked by clock_thread
					wait_mutex.lock();
					wait_mutex.unlock();
					if(r->poll() == false || global_stop){
						continue;
					}
					r->init_tick();
					com_ptr->adapter.send_daq(raster_index);
					break;
				}
			}
			r->init_tick();
			com_ptr->adapter.send_daq(raster_index);
			r->n_ticks++;
			while(global_stop == false){
				wait_mutex.lock();
				execute = false;
				// unlocked by clock_thread
				wait_mutex.lock();
				wait_mutex.unlock();
				if((r->is_sync == false && r->poll() == false) || global_stop){
					continue;
				}

				if(global_stop == false){
					r->tick();
					com_ptr->adapter.send_daq(raster_index);
					r->n_ticks++;
				}

			};
			r->final();
		}catch(abmt::util::bt_exception& b){
			cout << "An error occured in task " << r->name << ": " << b.what() << endl;
			b.print_backtrace();
			global_stop = true;
		}
	}

};



struct simulation_clock: public abmt::task{
	abmt::time start_time = 0;
	abmt::time stop_time = 0;
	abmt::time start_systime = 0;

	std::vector<std::shared_ptr<sim_task>> sync_tasks;
	std::vector<std::shared_ptr<sim_task>> async_tasks;

	simulation_clock():task([this]{this->run();}, false){

	}

	void run(){
		if(sync_tasks.size() == 0){
			return;
		}
		for(auto t:sync_tasks){
			t->n_ticks = 1;
		}

		std::vector<std::shared_ptr<sim_task>> tasks_waiting;
		while(global_stop == false){
			// poll all async tasks
			for(auto t:async_tasks){
				if(t->execute == false){
					t->execute = true;
					t->wait_mutex.unlock();
				}
			}
			// find next task
			tasks_waiting.clear();
			abmt::time min_time = sync_tasks[0]->next_start_time();
			tasks_waiting.push_back(sync_tasks[0]);
			for(int i = 1; i < sync_tasks.size(); ++i){
				auto task = sync_tasks[i];
				auto time = task->next_start_time();
				if(time < min_time){
					min_time = time;
					tasks_waiting.clear();
					tasks_waiting.push_back(task);
				}else if(time == min_time){
					tasks_waiting.push_back(task);
				}
			}

			// wait
			if(sim_speed != 0){
				auto real_delta = (get_systemtime() - start_systime).us();
				auto simulated_delta = (min_time - start_time).us();
				auto target_real_delta = simulated_delta/sim_speed;
				if(target_real_delta > real_delta){
					auto time_to_sleep = target_real_delta - real_delta;
					while(time_to_sleep > 0 && global_stop == false){
						if(time_to_sleep > 500*1000){
							usleep(500*1000);
							time_to_sleep -= 500*1000;
						}else{
							usleep(time_to_sleep);
							time_to_sleep = 0;
						}
					}

				}
			}
			// wait for task to be ready
			bool all_tasks_ready = false;
			while(all_tasks_ready == false && global_stop == false){
				all_tasks_ready = true;
				for(auto t:tasks_waiting){
					if(t->execute == true){
						all_tasks_ready = false;
					}
				}
			}

			// wait for async task to be ready
			all_tasks_ready = false;
			while(all_tasks_ready == false && global_stop == false){
				all_tasks_ready = true;
				for(auto t:async_tasks){
					if(t->execute == true){
						all_tasks_ready = false;
					}
				}
			}

			sim_time = min_time;
			for(auto t:tasks_waiting){
				t->execute = true;
				t->n_ticks++;
				t->wait_mutex.unlock();
			}

			if(sim_time >= stop_time){
				global_stop = true;
			}
		} // while global stop
		// start all tasks once again to let them quit properly
		for(auto t:sync_tasks){
			t->execute = true;
			t->wait_mutex.unlock();
		}
		for(auto t:async_tasks){
			t->execute = true;
			t->wait_mutex.unlock();
		}
	}
};

json read_json(string filename){
	ifstream f(filename);
	stringstream ss;
	ss << f.rdbuf();
	f.close();
	return json::parse(ss.str());
}



int main(){
	signal(SIGHUP,on_sig);
	signal(SIGINT,on_sig);
	signal(SIGQUIT,on_sig);

	std::setlocale(LC_ALL, "C.UTF-8");

	auto model_ptr = abmt::rt::get_model();

	abmt::io::event_list e;
	model_com_eio mdl_com(e);
	com_ptr = &mdl_com;
	mdl_com.set_model(model_ptr);
	mdl_com.adapter.on_quit = [&](){
		global_stop = true;
	};
	mdl_com.adapter.on_save_parameters = [&](){
		abmt::json res;
		for(auto p: model_ptr->parameters){
			std::string key = p.name;
			res[p.name] = abmt::serialize::to_json(p.parameter->get_type());
		}
		ofstream ofs("parameters.json");
		ofs << res.dump();
		ofs.close();
	};

	simulation_clock clk;
	abmt::time sys_start;
	bool time_from_file = false;


	// load config
	auto config = read_json("config.json");
	string start_time_str = config["start_time"];
	if(start_time_str.find("now") != string::npos){
		clk.start_time = get_systemtime();
	}else if(std::filesystem::exists(start_time_str)){
		auto p = std::filesystem::path(start_time_str);
		if(std::filesystem::is_regular_file(p) && (p.extension() == ".arec" || p.extension() == ".csv")){
				abmt::time start;
				abmt::time stop;
				if(p.extension() == ".arec"){
					start = get_start_time_from_arec(p.string());
					stop  = get_end_time_from_arec(p.string());
				}else if( p.extension() == ".csv"){
					start = get_start_time_from_csv(p.string());
					stop  = get_end_time_from_csv(p.string());
				}
				if(start != 0 && stop != 0){
						clk.start_time = start;
						clk.stop_time  = stop;
						time_from_file = true;
					}else{
						abmt::os::log_err("Unable to set time from recording/dir. Mabybe timestamps are missing. File/dir: '" + start_time_str + "'");
						exit(1);
				}
		}else{
			abmt::os::log_err("File in start_time is not a recording: '" + start_time_str + "'");
			exit(1);
		}
	}else{
		clk.start_time = abmt::util::str_read_date(start_time_str);
	}

	if(time_from_file == false){
		string stop = config["stop_time"];
		if(stop.find('-') != string::npos || stop.find(':') != string::npos || stop.find('/') != string::npos || stop.find('.') != string::npos){
			clk.stop_time = abmt::util::str_read_date(stop);
		}else{
			clk.stop_time = clk.start_time + ((int64_t)stod(stop))*1000LL*1000*1000;
		}
	}

	sim_speed = config["speed"];
	sim_time  = clk.start_time;

	// load parameters
	if(std::filesystem::exists("parameters.json")){
		json params = read_json("parameters.json");
		for(auto p: params){
			for(auto mdl_param: model_ptr->parameters){
				std::string name(mdl_param.name);
				if(name == p.key){
					auto exp = mdl_param.parameter->get_type();
					try{
						abmt::serialize::from_json(p.value,exp);
					}catch(...){
						std::cout << "error parsing parameter: " << name << std::endl;
					}
					break;
				}
			}
		}
	}
	mdl_com.adapter.on_save_parameters(); // update parameters for reproducibility

	// setup raster threads/sync_tasks
	std::map<std::string, std::shared_ptr<sim_task>> raster_tasks;

	for(size_t i = 0; i < model_ptr->rasters.length; ++i){
		auto r = model_ptr->rasters[i];
		auto task = make_shared<sim_task>(r,i);
		raster_tasks[r->name] = task;
		if(r->is_sync){
			clk.sync_tasks.push_back(task);
		}else{
			clk.async_tasks.push_back(task);
		}
		task->start();

	}


	try{

		abmt::os::die_if(clk.sync_tasks.size()==0,"No synchronous raster found to determine step width. Add delay...");
		auto raster_poll_time = clk.sync_tasks[0]->r->interval;
		for(auto& t: clk.sync_tasks){
			if(t->r->interval < raster_poll_time){
				raster_poll_time = t->r->interval;
			}
		}

		// delay start to give abmt time to connect and setup daq
		bool connect_delay = false;
		abmt::io::timer connect_timeout(e, 100, [&](){
			connect_delay = true;
		});
		while(connect_delay == false){
			e.wait(100);
		}

		// run
		cout << "Start time:           " << abmt::util::str_replace_date("#YYY-#M-#D #h:#m:#s", clk.start_time) << endl;
		cout << "Stop  time:           " << abmt::util::str_replace_date("#YYY-#M-#D #h:#m:#s", clk.stop_time) << endl;
		cout << "Async poll intervall: " << raster_poll_time.sec() << "s" << endl;
		cout << "Starting simulation..." << endl;
		clk.start_systime = get_systemtime();
		clk.start();
		while(global_stop == false){
			e.wait(100);
		}

		double duration = (get_systemtime() - clk.start_systime).sec();
		double sim_length = (sim_time - clk.start_time).sec();
		double speed = sim_length/duration;
		cout << "Simulation finished. Simulated  " << sim_length << "s in " << duration << "s. Speed: " <<  speed << "x" << endl;


		// quit
		bool all_tasks_ready = false;
		bool force_quit = false;
		abmt::io::timer quit_timer(e, 5000, [&](){ force_quit = true; });
		while(all_tasks_ready == false && force_quit == false){
			all_tasks_ready = true;
			for(auto t: raster_tasks){
				if(t.second->is_running()){
					all_tasks_ready = false;
					break;
				}
			}
			e.wait(100);
		}

		if(force_quit){
			cout << "Force exit. The following rasters are still running:" << endl;
			for(auto t: raster_tasks){
				if(t.second->is_running()){
					cout << t.first << endl;
				}
			}
			return 1;
		}

	}catch(abmt::util::bt_exception& b){
		cout << "An error occured in the main task: " << b.what() << endl;
		b.print_backtrace();
		global_stop = true;
	}

	return 0;
}
